/*
 * Copyright (c) 2016, Kelvin W Sherlock
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


#define XATTR_FINDERINFO_NAME "user.apple.FinderInfo"
#define XATTR_RESOURCEFORK_NAME "user.apple.ResourceFork"

#define XATTR_FILETYPE_NAME	"user.prodos.FileType"
#define XATTR_AUXTYPE_NAME	"user.prodos.AuxType"

#include "native_internal.h"

#include <unistd.h>
#include <fcntl.h>
#include <sys/xattr.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <utime.h>

#include <vector>
#include <algorithm>
#include <cstring>
#include <cerrno>

#include <macos/errors.h>

#ifndef ENOATTR
#define ENOATTR ENODATA
#endif

using MacOS::tool_return;
using MacOS::macos_error;
using MacOS::macos_error_from_errno;

namespace {

	class xattr_file final : public native::file {
	public:
		xattr_file(const std::string &path, int fd, bool readonly): file(path), _fd(fd), _readonly(readonly)
		{}
		~xattr_file(); 


		virtual tool_return<size_t> read(void *out_buffer, size_t count) override;
		virtual tool_return<size_t> write(const void *in_buffer, size_t count) override;
		virtual tool_return<size_t> get_mark() override;
		virtual tool_return<size_t> set_mark(ssize_t new_mark) override;
		virtual tool_return<size_t> get_eof() override;
		virtual tool_return<size_t> set_eof(ssize_t new_eof) override;


	private:

		ssize_t read_rfork(std::vector<uint8_t> &buffer);

		int _fd = -1;
		size_t _displacement = 0;
		bool _readonly = false;
	};

	xattr_file::~xattr_file() {
		close(_fd);
	}


	ssize_t xattr_file::read_rfork(std::vector<uint8_t> &buffer) {
		ssize_t eof = -1;
		buffer.clear();
		do {
			eof = fgetxattr(_fd, XATTR_RESOURCEFORK_NAME, NULL, 0);
			if (eof < 0) break;
			buffer.resize(eof);
			eof = fgetxattr(_fd, XATTR_RESOURCEFORK_NAME, buffer.data(), eof);
		} while (eof < 0 && errno == ERANGE);
		return eof;
	}

	tool_return<size_t> xattr_file::read(void *out_buffer, size_t count) {
		std::vector<uint8_t> buffer;
		ssize_t size;
		size = read_rfork(buffer);
		if (size < 0) size = 0;

		if (_displacement >= buffer.size()) return 0;

		count = std::min(count, buffer.size() - _displacement);
		std::copy_n(buffer.begin() + _displacement, count, (uint8_t *)out_buffer);
		_displacement += count;
		return count;
	}

	tool_return<size_t> xattr_file::write(const void *in_buffer, size_t count) {
		if (_readonly) return MacOS::wrPermErr;

		std::vector<uint8_t> buffer;
		ssize_t size;
		size = read_rfork(buffer);
		if (size < 0) size = 0;

		if (_displacement >= buffer.size()) {
			buffer.resize(_displacement, 0);
		}
		if (_displacement + count >= buffer.size()) {
			buffer.resize(_displacement + count, 0);
		}

		std::copy_n((const uint8_t *)in_buffer, count, buffer.begin() + _displacement);

		int ok = fsetxattr(_fd, XATTR_RESOURCEFORK_NAME, buffer.data(), buffer.size(), 0);
		if (ok < 0) return macos_error_from_errno();
		_displacement += count;
		return count;
	}



	tool_return<size_t> xattr_file::get_mark() {
		return _displacement;
	}

	tool_return<size_t> xattr_file::set_mark(ssize_t new_mark) {
		if (new_mark < 0) return MacOS::paramErr;
		_displacement = new_mark;
		return new_mark;
	}

	tool_return<size_t> xattr_file::get_eof() {
		ssize_t eof = fgetxattr(_fd, XATTR_RESOURCEFORK_NAME, NULL, 0);
		if (eof < 0) eof = 0;
		return eof;
	}

	tool_return<size_t> xattr_file::set_eof(ssize_t new_eof) {

		if (_readonly) return MacOS::wrPermErr;

		if (new_eof < 0) return MacOS::paramErr;

		std::vector<uint8_t> buffer;

		ssize_t eof = read_rfork(buffer);
		if (eof < 0) eof = 0;
		if (eof == new_eof) return MacOS::noErr;

		buffer.resize(new_eof, 0);
		int ok = fsetxattr(_fd, XATTR_RESOURCEFORK_NAME, buffer.data(), buffer.size(), 0);
		if (ok < 0) return macos_error_from_errno();
		return new_eof;
	}



	uint32_t rforksize(const std::string &path_name)
	{

		ssize_t ok = getxattr(path_name.c_str(), XATTR_RESOURCEFORK_NAME, NULL, 0);
		if (ok < 0) ok = 0;
		return ok;
	}



}
 
namespace native {


	macos_error get_finder_info(const std::string &path_name, void *info, bool extended) {

		uint8_t buffer[32];
		std::memset(buffer, 0, sizeof(buffer));

		ssize_t ok;
		ok = getxattr(path_name.c_str(), XATTR_FINDERINFO_NAME, buffer, 32);
		if (ok ==16 || ok == 32) {
			prodos_ftype_out(buffer);
			memcpy(info, buffer, extended ? 32 : 16);
			return MacOS::noErr;
		}

		if (ok < 0 && errno != ENOATTR && errno != ENOTSUP)  return macos_error_from_errno();

		/* if it's a text file, call it a text file */
		if (is_text_file_internal(path_name)) {
			memcpy(buffer, "TEXTMPS ", 8);
		}

		memcpy(info, buffer, extended ? 32 : 16);
		return MacOS::noErr;
	}


	macos_error set_finder_info(const std::string &path_name, const void *info, bool extended) {

		uint8_t buffer[32];
		ssize_t rv;

		std::memset(buffer, 0, sizeof(buffer));
		if (extended) {
			std::memcpy(buffer, info, 32);
		} else {
			std::memcpy(buffer, info, 16);
		}
		prodos_ftype_in(buffer);

		int ok = setxattr(path_name.c_str(), XATTR_FINDERINFO_NAME, buffer, 32, 0);
		if ( ok < 0) return macos_error_from_errno();

		return MacOS::noErr;
	}



	macos_error get_file_info(const std::string &path_name, file_info &fi)
	{
		struct stat st;

		if (stat(path_name.c_str(), &st) < 0)
			return macos_error_from_errno();

		fi.create_date = unix_to_mac(st.st_ctime);
		fi.modify_date = unix_to_mac(st.st_mtime);
		fi.backup_date = 0;

		if (S_ISDIR(st.st_mode)) {
			fi.type = file_info::directory;
			fi.attributes = 1 << 4;

			int links = st.st_nlink - 2;
			if (links < 0) links = 0;
			if (links > 65535) links = 65535;

			fi.entry_count = links;
			return MacOS::noErr;
		}

		// todo -- get actual block size instead of assuming 512.  oh well!
		
		fi.type = file_info::file;
		fi.data_logical_size = st.st_size;
		fi.data_physical_size = (st.st_size + 511) & ~511;
		fi.resource_physical_size = 0;
		fi.resource_logical_size = 0;

		get_finder_info(path_name, fi.finder_info);

		ssize_t rsize = rforksize(path_name);
		if (rsize > 0) {
			fi.resource_physical_size = rsize;
			fi.resource_logical_size = (rsize + 511) & ~511;
		}

		return MacOS::noErr;
	}


	macos_error set_file_info(const std::string &path_name, const file_info &fi) {

		struct stat st;

		if (stat(path_name.c_str(), &st) < 0) return macos_error_from_errno();

		if (S_ISREG(st.st_mode)) {
			auto rv = set_finder_info(path_name, fi.finder_info);
			if (rv) return rv;
		}

		time_t create_date = fi.create_date ? mac_to_unix(fi.create_date) : 0;
		time_t modify_date = fi.modify_date ? mac_to_unix(fi.modify_date) : 0;
		time_t backup_date = fi.backup_date ? mac_to_unix(fi.backup_date) : 0;

		// todo -- value of 0 == set to current date/time?

		if (modify_date == st.st_mtime) modify_date = 0;

		// try utimes.

		struct timeval tv[2];
		memset(tv, 0, sizeof(tv));

		int rv = 0;
		if (modify_date) {
			tv[0].tv_sec = st.st_atime;
			tv[1].tv_sec = modify_date;
			rv = utimes(path_name.c_str(), tv);
		}

		if (rv < 0) return macos_error_from_errno();
		return MacOS::noErr;
	}



	tool_return<file_ptr> open_resource_fork(const std::string &path_name, int oflag) {

		/* under HFS, every file has a resource fork.
		 * Therefore, create it if opening for O_RDWR or O_WRONLY
		 */

		int parent;
		if (oflag & O_CREAT) {
			int excl = oflag & O_EXCL;
			parent = open(path_name.c_str(), O_RDONLY | O_CREAT | excl, 0666);
		} else {
			parent = open(path_name.c_str(), O_RDONLY);
		}
		if (parent < 0) return macos_error_from_errno();

		int mode = oflag & O_ACCMODE;

		auto tmp = new xattr_file(path_name, parent, mode == O_RDONLY);
		tmp->resource = true;
		return file_ptr(tmp);
	}




}